import json
import shutil
import subprocess
from pathlib import Path

import typer
from rich.console import Console

from .config import CreateMode, Mode, Node, ViewMode
from .helper_main import create_dir, user_setted_ignores
from .tree.ignore import IgnoreManager
from .tree.renderer import FlatRenderer, JsonRenderer, TreeGenerator, TreeRenderer
from .tree.smd import dict_to_node, search_smd

app = typer.Typer(
    help=" Show directory tree (like 'tree') with smart ignore options and depth control."
)


@app.command()
def create_tree(
    smd_path: Path,
    parent: Path | None = None,
    dry_run: bool = typer.Option(False, "--dry-run", help="Simulate creation"),
    mode: CreateMode = typer.Option(  # noqa: B008
        CreateMode.SKIP,
        "--mode",
        help="File creation mode: skip (default), ask, or force overwrite",
    ),
    #
    # --- IGNORES ---
    ignore: list[str] = typer.Option(  # noqa: B008
        [],
        "--ignore",
        "-I",
        help="Ignore files or folders (comma-separated or multiple use)."
        " Example: -I '*.log,node_modules'",
    ),
    uv: bool = typer.Option(
        False, "--uv", help="Apply common Python/uv ignore patterns (.venv, pyproject.toml, etc.)"
    ),
    no_default_ignores: bool = typer.Option(
        False,
        "--no-default-ignores",
        help="Disable built-in ignore list (node_modules, .git, etc.)",
    ),
    python: bool = typer.Option(False, "--python-ignore", help="Ignore Python-related files"),
    node: bool = typer.Option(False, "--node-ignore", help="Ignore Node.js-related files"),
    git: bool = typer.Option(False, "--git-ignore", help="Ignore Git-related files"),
    editor: bool = typer.Option(False, "--editor-ignore", help="Ignore editor/OS files"),
    docker: bool = typer.Option(False, "--docker-ignore", help="Ignore Docker-related files"),
    rust: bool = typer.Option(False, "--rust-ignore", help="Ignore Rust-related files"),
):

    parent = parent or Path.cwd()
    smd_path = smd_path.resolve()
    console = Console()

    if not smd_path.exists() or not parent.exists():
        console.print("❌ File/Parent not found", style="bold red")
        raise typer.Exit(code=1)

    with smd_path.open("rt") as f:
        data = json.load(f)

    if data.get("_format") != "showmydir-v1":
        console.print(f"[bold red]❌ Unsupported format:[/bold red] {data.get('_format')}")
        raise typer.Exit(1)

    # Ignores the final
    final_ignore = user_setted_ignores(
        no_default_ignores=no_default_ignores,
        uv=uv,
        python=python,
        node=node,
        git=git,
        editor=editor,
        docker=docker,
        rust=rust,
        to_file=None,
    )
    if ignore:
        for item in ignore:
            for part in item.split(","):
                final_ignore.add(part.strip())

    root_node: Node = dict_to_node(data.get("root"))

    while True:
        answer = input("⚠️ This will create files. Continue? (y/n):") if not dry_run else "y"
        answer = answer.strip().lower()

        if answer == "y":
            console.print(
                "[red][DRY RUN MODE][/red]: (no files will be created)"
            ) if dry_run else ""
            create_dir(
                root_node=root_node,
                parent=parent,
                dry_run=dry_run,
                console=console,
                mode=mode,
                final_ignore=final_ignore,
            )
            break
        if answer == "n":
            typer.echo("Process canceled by user")
            break
        typer.echo("Invalid choice: ❌")


@app.command(help="Installs the .smd extension in vscode")
def install_showmydir_extension():
    vsix_path = Path(__file__).parent / "showmydir-syntax-0.0.1.vsix"
    code_cli = shutil.which("code")  # Only allow existing code CLI

    if not code_cli:
        typer.echo("⚠️ VS Code CLI not found in PATH")
        return

    subprocess.run([code_cli, "--install-extension", str(vsix_path)], check=True)  # noqa: S603
    print("✅ ShowMyDir syntax installed!")


@app.command()
def view(
    file_: Path,
    mode: ViewMode = typer.Option(ViewMode.TREE, "--mode", help="Mode: tree, flat, json"),  # noqa: B008
    show_size: bool = typer.Option(True, "--show-size/--no-show-size"),
    show_disk_size: bool = typer.Option(True, "--show-disk-size/--no-show-disk-size"),
    find: str = typer.Option(None, "--search", "-sch", help="Search .smd/.json file"),
):
    if not file_.exists():
        typer.echo("No such file found")
        raise typer.Exit(code=1)

    with file_.open("rt") as f:
        data = json.load(f)

    if data.get("_format") != "showmydir-v1":
        typer.echo(f"Unsupported format: {data.get('_format')}", err=True)
        raise typer.Exit(1)

    root_dict = data["root"]
    root_node = dict_to_node(root_dict)

    if find:
        match_nodes = search_smd(root_node, find)
        flat = FlatRenderer(
            root=file_,
            pathspecs=IgnoreManager(),
            depth=None,
            show_size=show_size,
            show_size_disk=show_disk_size,
            to_file=None,
            highlighted_files=set(),
        )
        for match in match_nodes:
            flat.render(match, Path(file_), find_mode=True)
    elif mode == ViewMode.JSON:
        typer.echo(json.dumps(data, indent=2))
    elif mode == ViewMode.TREE:
        TreeRenderer(show_size=show_size, show_disk=show_disk_size).render(root_node)
    elif mode == ViewMode.FLAT:
        flat = FlatRenderer(
            root=file_,
            depth=None,
            show_size=show_size,
            show_size_disk=show_disk_size,
            to_file=None,
            highlighted_files=set(),
            pathspecs=IgnoreManager(),
        )
        flat.render(
            node=root_node,
        )


@app.command()
def tree(
    # ignores
    ignore: list[str] = typer.Option(  # noqa: B008
        [],
        "--ignore",
        "-I",
        help="Ignore files or folders (comma-separated or multiple use)."
        " Example: -I '*.log,node_modules'",
    ),
    uv: bool = typer.Option(
        False,
        "--uv-ignore",
        help="Apply common Python/uv ignore patterns (.venv, pyproject.toml, etc.)",
    ),
    no_default_ignores: bool = typer.Option(
        False,
        "--no-default-ignores",
        help="Disable built-in ignore list (node_modules, .git, etc.)",
    ),
    python: bool = typer.Option(False, "--python-ignore", help="Ignore Python-related files"),
    node: bool = typer.Option(False, "--node-ignore", help="Ignore Node.js-related files"),
    git: bool = typer.Option(False, "--git-ignore", help="Ignore Git-related files"),
    editor: bool = typer.Option(False, "--editor-ignore", help="Ignore editor/OS files"),
    docker: bool = typer.Option(False, "--docker-ignore", help="Ignore Docker-related files"),
    rust: bool = typer.Option(False, "--rust-ignore", help="Ignore Rust-related files"),
    depth: int | None = typer.Option(
        None, "--depth", "-d", help="Limit recursion depth (e.g., -d 2)"
    ),
    root: Path | None = typer.Option(  # noqa: B008
        None,
        "--root",
        "-r",
        "--path",
        help="Root directory (default: current directory)",
    ),
    show_size: bool = typer.Option(
        True,
        "--show-size/--no-show-size",
        help="Toggle file size display",
    ),
    show_disk_size: bool = typer.Option(
        False, "--show-disk-size/--no-show-disk-size", help="Toggle dir disk size display"
    ),
    mode_of_viewing: Mode = typer.Option(  # noqa: B008
        Mode.TREE,
        "--mode",
        help="Mode: tree (default), flat, json",
    ),
    git_ignore: bool = typer.Option(
        True,
        "--gitignore/--no-gitignore",
        help="Reads gitignore files and excludes those files",
    ),
    highlights: list[str] = typer.Option(  # noqa: B008
        [],
        "--highlight",
        help="Highlight files matching patterns (e.g. '*.py', 'main.py')",
    ),
    file_saver: str = typer.Option(
        None,
        "--output",
        "-o",
        help="save the tree into file.",
    ),
):
    final_ignore = set()
    highlighted_set = set() if not highlights else set(highlights)

    # Checks the correct root path
    operation_dir = (root or Path.cwd()).resolve()
    if not operation_dir.exists():
        typer.echo(f"PathNotFoundError: {operation_dir} not exists")
        raise typer.Exit(code=1)

    # PathSpecs Ignores
    pt_specs_ig = IgnoreManager()
    pt_specs_ig.manage_pattern_list(
        no_default_ignores=no_default_ignores,
        uv=uv,
        python=python,
        node=node,
        git=git,
        editor=editor,
        docker=docker,
        rust=rust,
        to_file=file_saver,
    )
    pt_specs_ig.add_pattern(ignore)
    if git_ignore:
        pt_specs_ig.load_gitignore_file(operation_dir)
    pt_specs_ig.compile()

    # if user stated some ignores files/dirs
    if ignore:
        for item in ignore:
            for part in item.split(","):
                final_ignore.add(part.strip())

    # Mode of viewing
    if mode_of_viewing == Mode.TREE:
        TreeGenerator(
            root=operation_dir,
            ignore=pt_specs_ig,
            depth=depth,
            show_size=show_size,
            show_size_disk=show_disk_size,
            to_file=file_saver,
            highlighted_files=highlighted_set,
        )
    elif mode_of_viewing == Mode.FLAT:
        FlatRenderer(
            root=operation_dir,
            depth=depth,
            show_size=show_size,
            show_size_disk=show_disk_size,
            to_file=file_saver,
            highlighted_files=highlighted_set,
            pathspecs=pt_specs_ig,
        ).call_render(operation_dir)

    elif mode_of_viewing == Mode.JSON:
        file_name = Path(file_saver) if file_saver else None
        if file_name and file_name.suffix not in [".smd", ".json"]:
            typer.echo("JSON mode requires .smd or .json file", err=True)
            raise typer.Exit(1)

        JsonRenderer(
            root=operation_dir,
            ignore=pt_specs_ig,
            depth=depth,
            show_size=show_size,
            show_size_disk=show_disk_size,
            to_file=file_saver,
            highlighted_files=highlighted_set,
        )

    elif mode_of_viewing == Mode.SUMMARY:
        typer.echo("summary mode coming soon...")


# @app.callback()
def terminal_user_interface():
    pass


def main():
    app(prog_name="showmydir")


if __name__ == "__main__":
    main()



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import TypedDict


class Mode(str, Enum):
    TREE = "tree"
    FLAT = "flat"
    JSON = "json"
    SUMMARY = "summary"


class CreateMode(str, Enum):
    ASK = "ask"
    SKIP = "skip"
    FORCE = "force"


class ViewMode(str, Enum):
    FLAT = "flat"
    JSON = "json"
    TREE = "tree"


colour = str
Bytes = int


@dataclass
class Node:
    """
    Represents a file or directory in the tree.

    Attributes:
        name: Name of the file/directory
        is_dir: True if directory, False if file
        size: Actual file size (None if unknown)
        has_unknown_size: Whether size could not be determined
        logical_size: Aggregated size (used for directories)
        is_highlighted: Whether node matches highlight patterns
        children: List of child nodes (for directories)
    """

    name: str
    is_dir: bool
    size: Bytes | None
    has_unknown_size: bool
    logical_size: Bytes
    is_highlighted: bool
    children: list[Node] = field(default_factory=list)


class NodeDict(TypedDict, total=True):
    """Dictionary representation of Node used for JSON serialization."""

    name: str
    is_dir: bool
    size: int | None
    has_unknown_size: bool
    logical_size: int
    is_highlighted: bool
    children: list[NodeDict]


@dataclass
class MetaData:
    """Stores basic metadata for a file or directory."""

    name: str
    size: Bytes | None
    is_dir: bool


@dataclass
class FormatedSize:
    """Represents a human-readable file size with display color."""

    size: str
    colour: colour


# ===============================
# Default Ignore Groups
# ===============================


# Common directories/files ignored by default
DEFAULT_IGNORE_LIST = [
    "node_modules",
    ".git",
    "__pycache__",
    "bin",
    "dist",
    "*_venv",
    ".venv",
    "*.egg-info",
    ".pytest_cache",
    ".mypy_cache",
    ".DS_Store",
    ".idea",
    ".vscode",
]

UV_IGNORE_LIST = [
    ".venv",
    "pyproject.toml",
    "uv.lock",
    "uv.toml",
    ".gitignore",
    "README.md",
    ".python-version",
    ".env",
    "requirements.txt",
    "poetry.lock",
]

# Python-specific
PYTHON_IGNORES = [
    "__pycache__",
    "*.pyc",
    "*.pyo",
    "*.pyd",
    ".Python",
    "pip-wheel-metadata",
    "dist",
    "build",
    "*.egg-info",
    ".pytest_cache",
    ".mypy_cache",
    ".tox",
    ".coverage",
    "htmlcov",
]

# Node.js-specific
NODE_IGNORES = [
    "node_modules",
    "npm-debug.log",
    "yarn-error.log",
    "package-lock.json",
    "yarn.lock",
    ".pnpm-store",
]

# Git-specific
GIT_IGNORES = [
    ".git",
    ".gitignore",
    ".gitattributes",
    ".gitmodules",
]

# Editor / OS-specific
EDITOR_IGNORES = [
    ".DS_Store",
    "Thumbs.db",
    ".idea",
    ".vscode",
    "*.swp",
    "*.swo",
    "*~",
    ".directory",
    ".Spotlight-V100",
    ".Trashes",
]

# Docker / container related
DOCKER_IGNORES = [
    "Dockerfile",
    "docker-compose.yml",
    ".dockerignore",
    "*.tar",
    "*.tar.gz",
]

# Rust / Cargo
RUST_IGNORES = [
    "target",
    "Cargo.lock",
    "Cargo.toml",
]



from __future__ import annotations

import json
from importlib.metadata import version
from pathlib import Path
from typing import TYPE_CHECKING

import typer
from rich.console import Console
from rich.text import Text

from .scanner import ScanEngine
from .smd import node_to_dict
from .utils import choose_file_colour, format_node_size

if TYPE_CHECKING:
    from ..config import Node  # noqa: TID252
    from .ignore import IgnoreManager


class TreeGenerator(ScanEngine):
    def __init__(
        self,
        root: Path,
        ignore: IgnoreManager,
        depth: int | None,
        show_size: bool,
        show_size_disk: bool,
        to_file: str | None,
        highlighted_files: set[str],
    ):
        super().__init__(
            root=root,
            depth=depth,
            show_size=show_size,
            show_size_disk=show_size_disk,
            to_file=to_file,
            highlighted_files=highlighted_files,
            pathspecs=ignore,
        )

        self.root_node = self.scan(self.root)

        self.console.print(f"\n{self.root}", style="cyan")
        TreeRenderer(
            show_disk=self.show_disk, show_size=self.show_size, console=self.console
        ).render(node=self.root_node)
        self.show_denied_files()
        self._summary_stats()

        self._file_closer()


class TreeRenderer:
    def __init__(
        self, show_disk: bool = True, show_size: bool = True, console: Console | None = None
    ) -> None:
        self.show_disk = show_disk
        self.show_size = show_size
        self.console = console or Console()

    def render(self, node: Node | None, prefix: str = ""):
        if node is None:
            return

        for i, child in enumerate(node.children, start=1):
            is_last: bool = i == len(node.children)
            connector = "└── " if is_last else "├── "

            name = child.name + "/" if child.is_dir else child.name
            text = Text()
            text.append(prefix)
            text.append(connector)
            text.append(name, style=choose_file_colour(child.is_dir, child.is_highlighted))

            formatted_size = format_node_size(
                node=child, show_disk=self.show_disk, show_size=self.show_size
            )
            if formatted_size.size:
                text.append(f" ({formatted_size.size})", style=formatted_size.colour)

            self.console.print(text)

            new_prefix = prefix + ("    " if is_last else "│   ")
            if child.is_dir:
                self.render(child, new_prefix)


class JsonRenderer(ScanEngine):
    def __init__(
        self,
        root: Path,
        ignore: IgnoreManager,
        depth: int | None,
        show_size: bool,
        show_size_disk: bool,
        to_file: str | None,
        highlighted_files: set[str],
    ) -> None:
        super().__init__(root, depth, show_size, show_size_disk, to_file, highlighted_files, ignore)
        self.root_node = self.scan(self.root)
        if self.root_node is None:
            raise typer.Exit(code=1)
        data = {
            "_format": "showmydir-v1",
            "meta": {
                "generated_by": "showmydir",
                "version": version("showmydir"),
                "root_path": str(self.root),
                "total_files": self.file_counts,
                "total_dirs": self.dir_counts,
            },
            "root": node_to_dict(self.root_node),
        }
        json.dump(data, self.console.file, indent=2, ensure_ascii=False)

        self._file_closer()


class FlatRenderer(ScanEngine):
    def render(self, node: Node | None, parent_path: Path = Path(), find_mode: bool = False):
        if node is None:
            return

        current_path = parent_path / node.name
        relative_path = current_path.relative_to(self.root)

        # avoids printing root twice
        if node.name != self.root.name or parent_path != Path():
            text = Text()
            text.append(
                str(relative_path), style=choose_file_colour(node.is_dir, node.is_highlighted)
            )
            formated_size = format_node_size(
                node, show_size=self.show_size, show_disk=self.show_disk
            )
            text.append(f" ({formated_size.size})", style=formated_size.colour)
            self.console.print(text)

        if node.is_dir and not find_mode:
            for child in node.children:
                self.render(child, current_path)

    def call_render(self, dir: Path):
        nodes = self.scan(dir)
        self.render(nodes)



from __future__ import annotations

import fnmatch
import os
from pathlib import Path
from secrets import choice as secret_choice
from typing import TYPE_CHECKING

from rich.console import Console
from rich.style import Style
from rich.text import Text

# private files
from ..config import Bytes, MetaData, Node  # noqa: TID252
from .utils import format_bytes

if TYPE_CHECKING:
    from .ignore import IgnoreManager


class BaseGenerator:
    def __init__(
        self,
        root: Path,
        depth: int | None,
        show_size: bool,
        show_size_disk: bool,
        to_file: str | None,
        highlighted_files: set[str],
        pathspecs: IgnoreManager,
    ) -> None:
        self.root: Path = root.resolve()
        self.max_depth: int | None = depth
        self.show_size: bool = show_size
        self.show_disk: bool = show_size_disk
        self._file = to_file
        self.highlighted_files = highlighted_files

        self.dir_counts: int = 0
        self.file_counts: int = 0
        self.total_size: float = 0

        self.ignore = pathspecs

        self.console = (
            Console() if not self._file else Console(file=self._file_opener(Path(self._file)))
        )

        self.permission_error_files_dirs: list[Path] = []

    def _file_opener(self, __file: Path, mode: str = "w"):
        self.f = open(__file, mode)  # type: ignore  # noqa: PTH123, SIM115
        return self.f

    def _file_closer(self):
        if hasattr(self.console, "file") and self.console.file:
            self.console.file.close()

    def _get_entries(self, path: Path):
        try:
            entries = sorted(
                os.scandir(path),
                key=lambda e: (not e.is_dir(), e.name.lower()),
            )
            for entry in entries:
                # makes a relative to root dir so pathspecs work properly
                # as_posix make '\' into '/', to suppport windows too
                entry_path = Path(entry.path)
                rel_path = entry_path.relative_to(self.root).as_posix()
                if not self.ignore.should_include(rel_path):
                    continue
                yield entry

        except PermissionError:
            self.permission_error_files_dirs.append(path)
            return

    def _summary_stats(self):
        _size_data = format_bytes(bytes_size=self.total_size, dim=False)
        dir_size, colour_ = _size_data.size, _size_data.colour
        self.console.print()
        self.console.print(
            f"[bold]📊 {self.file_counts} files, {self.dir_counts} folders/directories[/bold]"
        )
        self.console.print()
        text = Text("💾 Total size: ", style="bold")
        text.append(dir_size, style=colour_)
        self.console.print(text)

    def show_denied_files(self):
        if self.permission_error_files_dirs:
            self.console.print("\n[Permission Denied for these files/folders]:", style="bold red")

            # Pick one color randomly for this run
            dim_colors = ["grey50", "grey62", "grey70", "light_slate_gray", "grey35", "grey42"]
            chosen_color = secret_choice(dim_colors)
            style = Style(color=chosen_color)

            for path in self.permission_error_files_dirs:
                text = Text(f"  • {path}", style=style)
                self.console.print(text)

    def _is_highlighted(self, name: str) -> bool:
        return any(fnmatch.fnmatch(name, pattern) for pattern in self.highlighted_files)

    def _extract_metadata(self, entry: Path, stat: os.stat_result | None) -> MetaData:
        if stat is None:
            return MetaData(entry.name, None, False)

        if entry.is_dir():
            return MetaData(entry.name, stat.st_size, True)

        return MetaData(entry.name, stat.st_size, False)


class ScanEngine(BaseGenerator):
    def __init__(
        self,
        root: Path,
        depth: int | None,
        show_size: bool,
        show_size_disk: bool,
        to_file: str | None,
        highlighted_files: set[str],
        pathspecs: IgnoreManager,
    ) -> None:
        super().__init__(
            root,
            depth,
            show_size,
            show_size_disk,
            to_file,
            highlighted_files,
            pathspecs,
        )

    def scan(self, path: Path, current_depth: int = 0) -> Node | None:
        if self.max_depth is not None and current_depth >= self.max_depth:
            return None

        try:
            stat = path.stat()
        except (
            PermissionError,
            FileNotFoundError,
        ):  # some files change thier path while scanning
            stat = None

        _path_data = self._extract_metadata(entry=path, stat=stat)
        dir_file_name, size, is_dir = _path_data.name, _path_data.size, _path_data.is_dir
        highlighted = self._is_highlighted(path.name)

        has_unknown_size: bool = size is None
        logical_size: Bytes = size if not is_dir and size is not None else 0

        node = Node(
            name=dir_file_name,
            is_dir=is_dir,
            size=size,
            has_unknown_size=has_unknown_size,
            logical_size=logical_size,
            is_highlighted=highlighted,
        )

        if is_dir:
            if current_depth > 0:
                self.dir_counts += 1
            for entry in self._get_entries(path):
                child_node = self.scan(Path(entry.path), current_depth=current_depth + 1)
                if child_node is not None:
                    node.children.append(child_node)
                    node.logical_size += child_node.logical_size

                    if child_node.has_unknown_size:
                        node.has_unknown_size = True
        else:
            self.file_counts += 1
            self.total_size += node.size or 0

        return node



from __future__ import annotations

import fnmatch

from ..config import Node, NodeDict  # noqa: TID252


def node_to_dict(node: Node) -> NodeDict:
    """
    Converts a Node object into a dictionary representation.

    This is used for serializing the tree into JSON or `.smd` format.
    """
    return {
        "name": node.name,
        "is_dir": node.is_dir,
        "size": node.size,
        "has_unknown_size": node.has_unknown_size,
        "logical_size": node.logical_size,
        "is_highlighted": node.is_highlighted,
        "children": [node_to_dict(child) for child in node.children],
    }


def dict_to_node(data: NodeDict) -> Node:
    """
    Reconstructs a Node object from its dictionary representation.

    Used when loading `.smd` or JSON files back into a tree structure.
    """
    return Node(
        name=data["name"],
        is_dir=data["is_dir"],
        size=data["size"],
        has_unknown_size=data["has_unknown_size"],
        logical_size=data["logical_size"],
        is_highlighted=data["is_highlighted"],
        children=[dict_to_node(child) for child in data["children"]],
    )


def search_smd(root_node: Node, pattern: str) -> list[Node]:
    matches: list[Node] = []

    def _search(node: Node):
        if fnmatch.fnmatch(node.name, pattern):
            matches.append(node)
        for child in node.children:
            _search(child)

    _search(root_node)
    return matches



class CreateModeNotUpdatedError(Exception):
    """Raised when a new CreateMode is added but not handled in logic."""

    def __init__(self, message: str | None = None):
        super().__init__(message or "Program is not updated here.")


class GitIgnoreFileNotFoundError(Exception):
    """Raised when a .gitignore file is requested but not found."""

    def __init__(self, message: str | None = None) -> None:
        super().__init__(message or "No such gitignore file found")


class InvalidNodeNameError(Exception):
    def __init__(self, name: str, message: str | None = None) -> None:
        super().__init__(message or f"Invalid node name: {name}")


class IgnoreManagerCompileNotCalledError(Exception):
    def __init__(self, message: str | None = None) -> None:
        super().__init__(message or "IgnoreManager not compiled. Call compile() first.")



from ..config import Bytes, FormatedSize, Node, colour  # noqa: TID252


def format_bytes(bytes_size: Bytes | float | None, dim: bool = True) -> FormatedSize:
    if bytes_size is None:  # for permission errors
        return FormatedSize("no access", "red")
    if bytes_size < 1024:
        return FormatedSize(f"{bytes_size} B", "dim white" if dim else "white")
    if bytes_size < 1_048_576:
        return FormatedSize(f"{bytes_size / 1024:.2f} KB", "dim green" if dim else "green")
    if bytes_size < 1_073_741_824:
        return FormatedSize(f"{bytes_size / 1_048_576:.2f} MB", "dim yellow" if dim else "yellow")
    return FormatedSize(f"{bytes_size / 1_073_741_824:.2f} GB", "dim red" if dim else "red")


def choose_file_colour(is_dir: bool, is_highlighted: bool) -> colour:
    if is_highlighted:
        return "bold bright_blue"
    if is_dir:
        return "cyan"
    return "green"


def format_node_size(
    node: Node, show_size: bool, show_disk: bool, dim: bool = True
) -> FormatedSize:
    parts = []

    size_data = format_bytes(node.logical_size, dim)
    logical_str = size_data.size

    if show_size:  # if should show size
        if node.has_unknown_size:
            parts.append(f"{logical_str} (~)")  # AND system dont allow the actual size of file/dir
        else:
            parts.append(logical_str)  # AND system allow to get size of dir/file

    # for directories, also show disk size if not false
    if node.is_dir:
        disk_data = format_bytes(node.size, dim)
        disk_size, disk_colour = disk_data.size, disk_data.colour
        if show_disk:
            parts.append(f"Disk: {disk_size}")
        return FormatedSize((" | ").join(parts), (size_data.colour if show_disk else disk_colour))

    # Now here only files remains
    if show_size:
        return FormatedSize(logical_str, size_data.colour)
    return FormatedSize("", "")



from pathlib import Path

from pathspec import PathSpec

from .. import config  # noqa: TID252
from ..customExceptions import IgnoreManagerCompileNotCalledError  # noqa: TID252


class IgnoreManager:
    def __init__(self, pattens: list[str] | None = None) -> None:
        self.ignore_list = pattens or []

    def compile(self):
        self.spec = PathSpec.from_lines("gitwildmatch", self.ignore_list)

    def should_include(self, path: str) -> bool:
        if not hasattr(self, "spec"):
            raise IgnoreManagerCompileNotCalledError()
        return not self.spec.match_file(path)

    def load_gitignore_file(self, path: Path):
        path = path / ".gitignore"
        if not path.exists():
            return
        try:
            with path.open("r") as f:
                # read lines and add to list
                self.ignore_list.extend(
                    line.strip() for line in f if line.strip() and not line.strip().startswith("#")
                )
        except (PermissionError, FileNotFoundError):
            pass

    def add_pattern(self, pattern: list[str]):
        for item in pattern:
            self.ignore_list.extend(item.strip().split(","))

    def manage_pattern_list(
        self,
        no_default_ignores: bool,
        uv: bool,
        python: bool,
        node: bool,
        git: bool,
        editor: bool,
        docker: bool,
        rust: bool,
        to_file: str | None,
    ) -> list[str]:
        """
        Build the final ignore set based on user-selected options.

        Combines default ignore patterns with optional categories such as
        Python, Node.js, Git, Docker, etc. Also ensures the output file
        is excluded if specified.

        Args:
            no_default_ignores (bool): Disable default ignore patterns.
            uv (bool): Include uv-related ignores.
            python (bool): Include Python-related ignores.
            node (bool): Include Node.js-related ignores.
            git (bool): Include Git-related ignores.
            editor (bool): Include editor/OS-specific ignores.
            docker (bool): Include Docker-related ignores.
            rust (bool): Include Rust-related ignores.
            to_file (str | None): Output file name to ignore.

        Returns:
            list[str]: Final set of ignore patterns.
        """
        if not no_default_ignores:
            self.ignore_list.extend(config.DEFAULT_IGNORE_LIST)

        if uv:
            self.ignore_list.extend(config.UV_IGNORE_LIST)
        if python:
            self.ignore_list.extend(config.PYTHON_IGNORES)
        if node:
            self.ignore_list.extend(config.NODE_IGNORES)
        if git:
            self.ignore_list.extend(config.GIT_IGNORES)
        if editor:
            self.ignore_list.extend(config.EDITOR_IGNORES)
        if docker:
            self.ignore_list.extend(config.DOCKER_IGNORES)
        if rust:
            self.ignore_list.extend(config.RUST_IGNORES)

        if to_file:
            # if tree will genrate then it should also ignore that perticular file also
            # otherwise incomplete data may also show on tree
            self.ignore_list.append(Path(to_file).name)

        return self.ignore_list



import re
from pathlib import Path

from rich.console import Console

from . import config
from .config import CreateMode, Node
from .customExceptions import (
    CreateModeNotUpdatedError,
    GitIgnoreFileNotFoundError,
    InvalidNodeNameError,
)

# The string must start and end with one or more alphanumeric characters, underscores, hyphens, or periods.  # noqa: E501
SAFE_NAME_PATTERN = re.compile(r"^(?!.*\.{2})[\w\-.]+$")


def sanitize_name(name: str) -> str:
    """
    Sanitize a node name to prevent path traversal or unsafe characters.

    Only allows: a-z, A-Z, 0-9, _, -, .
    Rejects names containing slashes or '..'.
    """
    if not SAFE_NAME_PATTERN.match(name) or not name.strip():
        raise InvalidNodeNameError(name)
    return name


def handle_file(
    safe_name: str,
    parent: Path,
    mode: CreateMode,
    dry_run: bool,
    console: Console,
    indent: str,
):
    """
    Handle creation or overwrite behavior for a single file node.

    Depending on the selected mode (ASK, SKIP, FORCE), this function
    decides whether to create, skip, or overwrite a file. It also
    supports dry-run mode where no actual changes are made.

    Args:
        node (Node): File node containing metadata.
        parent (Path): Parent directory path where file will be created.
        mode (CreateMode): File handling strategy (ASK, SKIP, FORCE).
        dry_run (bool): If True, simulate actions without creating files.
        console (Console): Rich console instance for output.
        indent (str): Indentation used for structured printing.

    Raises:
        CreateModeNotUpdatedError: If an unknown mode is provided.
    """

    if mode == CreateMode.ASK:
        message = "[PROMPT]"
    elif mode == CreateMode.SKIP:
        message = "[SKIP]"
    elif mode == CreateMode.FORCE:
        message = "[FORCED]"

    current_path = parent / safe_name

    if current_path.exists():
        if dry_run:
            console.print(
                f"{indent}[red]{message}[/red] [green]{safe_name}[/green] "
                + ("(would ask)" if mode == CreateMode.ASK else "")
            )
        else:
            if mode == CreateMode.ASK:
                while True:
                    choice = (
                        console.input(
                            "[red]⚠️[/red] Conflicting file found, want to overwrite? [y/n]: "
                        )
                        .strip()
                        .lower()
                    )
                    if choice == "y":
                        current_path.write_text("")
                        break
                    if choice == "n":
                        break
                    console.print("Invalid choice...")
            elif mode == CreateMode.FORCE:
                current_path.touch()
            elif mode == CreateMode.SKIP:
                console.print(f"Successfully Skipped file: {safe_name}")
            else:
                raise CreateModeNotUpdatedError()
    else:
        if dry_run:
            console.print(f"{indent}[green]{safe_name}[/green]")
        else:
            current_path.touch()


def handle_dir(
    safe_name: str,
    parent: Path,
    mode: CreateMode,
    dry_run: bool,
    console: Console,
    indent: str = "   ",
):
    """
    Handle creation or overwrite behavior for a directory node.

    Creates directories or simulates creation based on the selected
    mode and dry-run setting. Unlike files, directories typically do
    not require overwrite confirmation.

    Args:
        node (Node): Directory node.
        parent (Path): Parent directory path.
        mode (CreateMode): Directory handling strategy.
        dry_run (bool): If True, simulate actions without creating directories.
        console (Console): Rich console instance for output.
        indent (str): Indentation used for structured printing.

    Raises:
        CreateModeNotUpdatedError: If an unknown mode is provided.
    """

    if mode == CreateMode.ASK:
        message = "[PROMPT]"
    elif mode == CreateMode.SKIP:
        message = "[SKIP]"
    elif mode == CreateMode.FORCE:
        message = "[OVERWRITE]"

    current_path = parent / safe_name

    if current_path.exists():
        if dry_run:
            console.print(
                f"{indent}[red]{message}[/red] [cyan]{safe_name}[/cyan] "
                + ("(would ask)" if mode == CreateMode.ASK else "")
            )
        else:
            if mode == CreateMode.ASK:
                # In both cases (overwrites or not) directory will always show no effect
                # dir no matter what/how, it will got created
                pass  # directories don't need overwrite confirmation
            elif mode == CreateMode.FORCE:
                current_path.mkdir(parents=True, exist_ok=True)
            elif mode == CreateMode.SKIP:
                console.print(f"Successfully Skipped directory: {safe_name}")
            else:
                raise CreateModeNotUpdatedError()
    else:
        if dry_run:
            console.print(f"{indent}[cyan]{safe_name}[/cyan]")
        else:
            current_path.mkdir(parents=True, exist_ok=True)


def create_dir(
    root_node: Node,
    parent: Path,
    dry_run: bool,
    mode: CreateMode,
    console: Console,
    final_ignore: set[str],
    level: int = 0,
    safe_dir_name: str | None = None,
):
    """
    Recursively create directory and file structure from a Node tree.

    Traverses the given Node structure and creates directories and files
    based on the selected mode. Supports dry-run to preview changes
    without modifying the filesystem.

    Args:
        root_node (Node): Root node of the directory structure.
        parent (Path): Base path where structure will be created.
        dry_run (bool): If True, simulate actions without creating files.
        mode (CreateMode): File/directory handling strategy.
        console (Console): Rich console instance for output.
        final_ignore (set[str]): Set of file/directory names to ignore.
        level (int, optional): Current recursion depth for indentation.

    Returns:
        None
    """
    try:
        safe_name = safe_dir_name or sanitize_name(root_node.name)
        current_path = parent / safe_name
        indent = "   " * level
        if root_node.is_dir:
            handle_dir(safe_name, parent, mode, dry_run, console, indent)

        for node in root_node.children:
            child_indent = "   " * (level + 1)
            try:
                child_safe_name = sanitize_name(node.name)
            except InvalidNodeNameError:
                continue
            if node.name in final_ignore:
                continue  # ignores the child who are in final ignore
            if node.is_dir:
                create_dir(
                    root_node=node,
                    parent=current_path,
                    dry_run=dry_run,
                    console=console,
                    mode=mode,
                    final_ignore=final_ignore,
                    level=level + 1,
                    safe_dir_name=child_safe_name,
                )
            else:
                handle_file(
                    safe_name=child_safe_name,
                    parent=current_path,
                    mode=mode,
                    dry_run=dry_run,
                    console=console,
                    indent=child_indent,
                )

    except InvalidNodeNameError:
        # try/except chamer is this huge because it ignores whole dir if its name is not safe
        console.print(f"[red]InvalidNode:[/red] {root_node.name}")


def user_setted_ignores(
    no_default_ignores: bool,
    uv: bool,
    python: bool,
    node: bool,
    git: bool,
    editor: bool,
    docker: bool,
    rust: bool,
    to_file: str | None,
) -> set[str]:
    """
    Build the final ignore set based on user-selected options.

    Combines default ignore patterns with optional categories such as
    Python, Node.js, Git, Docker, etc. Also ensures the output file
    is excluded if specified.

    Args:
        no_default_ignores (bool): Disable default ignore patterns.
        uv (bool): Include uv-related ignores.
        python (bool): Include Python-related ignores.
        node (bool): Include Node.js-related ignores.
        git (bool): Include Git-related ignores.
        editor (bool): Include editor/OS-specific ignores.
        docker (bool): Include Docker-related ignores.
        rust (bool): Include Rust-related ignores.
        to_file (str | None): Output file name to ignore.

    Returns:
        set[str]: Final set of ignore patterns.
    """

    final_ignore: set[str] = set()
    if not no_default_ignores:
        final_ignore.update(config.DEFAULT_IGNORE_LIST)

    if uv:
        final_ignore.update(config.UV_IGNORE_LIST)
    if python:
        final_ignore.update(config.PYTHON_IGNORES)
    if node:
        final_ignore.update(config.NODE_IGNORES)
    if git:
        final_ignore.update(config.GIT_IGNORES)
    if editor:
        final_ignore.update(config.EDITOR_IGNORES)
    if docker:
        final_ignore.update(config.DOCKER_IGNORES)
    if rust:
        final_ignore.update(config.RUST_IGNORES)

    if to_file:
        final_ignore.add(Path(to_file).name)

    return final_ignore


def read_gitignore(operation_dir: Path) -> list[str]:
    """
    Read and parse a .gitignore file into a list of ignore patterns.

    Ignores empty lines and comments, and trims trailing slashes
    from directory entries.

    Args:
        operation_dir (Path): Directory containing the .gitignore file.

    Raises:
        GitIgnoreFileNotFoundError: If .gitignore file is not found.

    Returns:
        list[str]: List of ignore patterns.
    """

    git_ignore_file = operation_dir / ".gitignore"

    if not git_ignore_file.exists():
        raise GitIgnoreFileNotFoundError()

    # if exists then opens it
    with git_ignore_file.open() as f:
        temp_ignore_list: list[str] = []  # container to store ignoring files/dir
        for line in f:
            line = line.strip()  # clean a line

            if line.lstrip().startswith("#"):
                continue  # ignores for comment
            ig_part = line.strip()  # remove unwanted

            if ig_part != "":
                temp_ignore_list.append(ig_part.rstrip("/"))

    return temp_ignore_list
